import ApplicationComponent from "../ApplicationComponent";
import {Express} from "express";
import redis, {RedisClient} from "redis";
import config from "config";
import Logger from "../Logger";
import session, {Store} from "express-session";
import connect_redis from "connect-redis";
import CacheProvider from "../CacheProvider";

const RedisStore = connect_redis(session);

export default class RedisComponent extends ApplicationComponent<void> implements CacheProvider {
    private redisClient?: RedisClient;
    private store?: Store;

    public async start(app: Express): Promise<void> {
        this.redisClient = redis.createClient(config.get('redis.port'), config.get('redis.host'), {
            password: config.has('redis.password') ? config.get<string>('redis.password') : undefined,
        });
        this.redisClient.on('error', (err: any) => {
            Logger.error(err, 'An error occurred with redis.');
        });
        this.store = new RedisStore({
            client: this.redisClient,
            prefix: config.get<string>('redis.prefix') + '-session:',
        });
    }

    public async stop(): Promise<void> {
        if (this.redisClient) {
            await this.close('Redis connection', this.redisClient, this.redisClient.quit);
        }
    }

    public getStore(): Store {
        if (!this.store) throw `Redis store was not initialized.`;
        return this.store;
    }

    public canServe(): boolean {
        return this.redisClient !== undefined && this.redisClient.connected;
    }

    public async get(key: string, defaultValue?: string): Promise<string | null> {
        return new Promise<string | null>((resolve, reject) => {
            if (!this.redisClient) {
                reject(`Redis store was not initialized.`);
                return;
            }

            this.redisClient.get(key, (err, val) => {
                if (err) {
                    reject(err);
                    return;
                }
                resolve(val !== null ? val : (defaultValue !== undefined ? defaultValue : null));
            });
        });
    }

    public async has(key: string): Promise<boolean> {
        return await this.get(key) !== null;
    }

    public async forget(key: string): Promise<void> {
        return new Promise<void>((resolve, reject) => {
            if (!this.redisClient) {
                reject(`Redis store was not initialized.`);
                return;
            }

            this.redisClient.del(key, (err) => {
                if (err) {
                    reject(err);
                    return;
                }

                resolve();
            });
        });
    }

    public async remember(key: string, value: string, ttl: number): Promise<void> {
        return new Promise<void>((resolve, reject) => {
            if (!this.redisClient) {
                reject(`Redis store was not initialized.`);
                return;
            }

            this.redisClient.psetex(key, ttl, value, (err) => {
                if (err) {
                    reject(err);
                    return;
                }
                resolve();
            });
        });
    }
}